/* setfilever.c - Add or update VERSIONINFO resource for the    */
/*                specified executable file.                    */

/* All text in the VERSIONINFO block is Unicode. Set it as default.  */
#define UNICODE

#include <windows.h>
#include <string.h>
#include <stdio.h>
#include "verinfo.h"


VerBuff VersionBuffer;
FixedInfo *pFixedInfo = NULL;
StringInfoBuff *stringInfo;

int gDiag = 0;
int gVerExists = 0;

HMODULE TargetFile;
HFILE ResFile;
wchar_t TargetName[MAX_PATH] = L"";
wchar_t fullTargetPath[MAX_PATH] = L"";
wchar_t *targetFileName = NULL;
wchar_t ResName[MAX_PATH]    = L"";
wchar_t MyName[MAX_PATH]     = L"";
wchar_t aCompanyName[MAX_PATH]      = L"";
wchar_t aCopyright[MAX_PATH]        = L"";
wchar_t aCompatibilityLevel[MAX_PATH] = L"";
wchar_t aFileDescription[MAX_PATH]  = L"";
wchar_t aInternalFileName[MAX_PATH] = L"";
wchar_t aOriginalFileName[MAX_PATH] = L"";
wchar_t aProductName[MAX_PATH]      = L"";
wchar_t aPrivateBuild[MAX_PATH]     = L"";
wchar_t aComments[MAX_PATH]         = L"";
wchar_t aSpecialBuild[MAX_PATH]     = L"";
wchar_t aTrademark[MAX_PATH]        = L"";
wchar_t aFileVersion[MAX_PATH]      = L"";
wchar_t aProductVersion[MAX_PATH]   = L"";



extern int resourceSectionExists(wchar_t *fileName);
extern int addResourceSection(wchar_t *fileName);


void displaySyntax(void)
{
	fprintf(stderr, "\nThe setfilever utility will add a Windows VERSIONINFO (English)\n"
		"resource or modify the existing one for a Win32 .exe or .dll file\n\n");

	fprintf(stderr, "Syntax: \n"
		"setfilever [-C <company name>] [-c <copyright>] [-F <file description]\n"
		"       [-I <internal file name>] [-L <Level>] [-N <original file name>]\n"
		"       [-P <product name>] [-p <private build] [-R <remarks or comments>]\n"
		"       [-S <special build>] [-T <trademark>] [-V <file version>]\n"
		"       [-v <product version>] target\n\n");

	fprintf(stderr, "Description of options:\n"
		"-C\t company name, default is empty\n"
		"-c\t copyright string, default is \"Copyright \157 2012\"\n"
		"-F\t file description, default is empty\n"
		"-I\t internal file name, default is empty\n"
		"-L\t component compatibility level (an integer), default is empty\n"
		"-N\t original file name, default is filename of target file\n"
		"-P\t product name, default is empty\n"
		"-p\t private build, default is empty\n"
		"-R\t remarks (comments), default is empty\n"
		"-S\t special build, default is empty\n"
		"-T\t trademark, default is empty\n"
		"-V\t file version, default is 1.0.0.0\n"
		"-v\t product version, default is 1.0.0.0\n\n");

	fprintf(stderr, "setfilever [-H] [-h] [-?] will display this help message.\n\n");

	fprintf(stderr, "NOTES:\n"
					"      This program will directly modify the specified file. It is\n"
					"      possible the file could be corrupted. Please save a backup\n"
					"      copy of your file before attempting to use this utility.\n\n"
		            "      If the target file already has a VERSIONINFO resource, default\n"
		            "      values will not be applied. Existing values will be preserved\n"
					"      unless explicitly overridden with new values on command line.\n\n"
					"      String values that contain spaces must be enclosed in quotes.\n\n"
					"      Version strings must contain 2 to 4 numeric fields.\n\n"
					"      If target is a device driver, use EDITBIN /RELEASE to recalculate\n"
					"      the checksum after running this utility.\n\n");
}


void getArgs(int argc, wchar_t **argv)
  {
	int i;
	wchar_t *sptr;

    wcscpy(MyName, argv[0]);

	if (!wcsstr(_wcslwr(MyName), L".exe"))
		wcscat(MyName, L".exe");

	if (argc < 2)
	{
		displaySyntax();
		exit(0);
	}

	for (i = 1; i < argc; i++)
	{
		if (*argv[i] == L'-')
		{
			switch(argv[i][1])
			{
			case L'C':
				if (!argv[i][2])
				{
					if (++i >= argc)
					{
						fprintf(stderr, "ERROR: Invalid argument for %S\n\n", argv[i-1]);
						displaySyntax();
				                exit(1);
					}

					sptr = argv[i];
				}
				else
					sptr = &argv[i][2];

			    wcsncpy(aCompanyName, sptr, MAX_PATH);
				aCompanyName[MAX_PATH - 1] = 0;
                break;

			case L'c':
				if (!argv[i][2])
				{
					if (++i >= argc)
					{
						fprintf(stderr, "ERROR: Invalid argument for %S\n\n", argv[i-1]);
						displaySyntax();
		                exit(1);
					}

					sptr = argv[i];
				}
				else
					sptr = &argv[i][2];

			    wcsncpy(aCopyright, sptr, MAX_PATH);
				aCopyright[MAX_PATH - 1] = 0;
                break;

			case L'D':
			case L'd':
				gDiag = 1;
                break;

			case L'F':
				if (!argv[i][2])
				{
					if (++i >= argc)
					{
						fprintf(stderr, "ERROR: Invalid argument for %S\n\n", argv[i-1]);
						displaySyntax();
		                exit(1);
					}

					sptr = argv[i];
				}
				else
					sptr = &argv[i][2];

			    wcsncpy(aFileDescription, sptr, MAX_PATH);
				aFileDescription[MAX_PATH - 1] = 0;
                break;

			case L'I':
				if (!argv[i][2])
				{
					if (++i >= argc)
					{
						fprintf(stderr, "ERROR: Invalid argument for %S\n\n", argv[i-1]);
						displaySyntax();
		                exit(1);
					}

					sptr = argv[i];
				}
				else
					sptr = &argv[i][2];

			    wcsncpy(aInternalFileName, sptr, MAX_PATH);
				aInternalFileName[MAX_PATH - 1] = 0;
                break;

			case L'L':
				if (!argv[i][2])
				{
					if (++i >= argc)
					{
						fprintf(stderr, "ERROR: Invalid argument for %S\n\n", argv[i-1]);
						displaySyntax();
		                exit(1);
					}

					sptr = argv[i];
				}
				else
					sptr = &argv[i][2];

			    wcsncpy(aCompatibilityLevel, sptr, MAX_PATH);
				aCompatibilityLevel[MAX_PATH - 1] = 0;
                break;

			case L'N':
				if (!argv[i][2])
				{
					if (++i >= argc)
					{
						fprintf(stderr, "ERROR: Invalid argument for %S\n\n", argv[i-1]);
						displaySyntax();
		                exit(1);
					}

					sptr = argv[i];
				}
				else
					sptr = &argv[i][2];

			    wcsncpy(aOriginalFileName, sptr, MAX_PATH);
				aOriginalFileName[MAX_PATH - 1] = 0;
                break;

			case L'P':
				if (!argv[i][2])
				{
					if (++i >= argc)
					{
						fprintf(stderr, "ERROR: Invalid argument for %S\n\n", argv[i-1]);
						displaySyntax();
		                exit(1);
					}

					sptr = argv[i];
				}
				else
					sptr = &argv[i][2];

			    wcsncpy(aProductName, sptr, MAX_PATH);
				aProductName[MAX_PATH - 1] = 0;
                break;

			case L'p':
				if (!argv[i][2])
				{
					if (++i >= argc)
					{
						fprintf(stderr, "ERROR: Invalid argument for %S\n\n", argv[i-1]);
						displaySyntax();
		                exit(1);
					}

					sptr = argv[i];
				}
				else
					sptr = &argv[i][2];

			    wcsncpy(aPrivateBuild, sptr, MAX_PATH);
				aPrivateBuild[MAX_PATH - 1] = 0;
                break;

			case L'R':
				if (!argv[i][2])
				{
					if (++i >= argc)
					{
						fprintf(stderr, "ERROR: Invalid argument for %S\n\n", argv[i-1]);
						displaySyntax();
		                exit(1);
					}

					sptr = argv[i];
				}
				else
					sptr = &argv[i][2];

			    wcsncpy(aComments, sptr, MAX_PATH);
				aComments[MAX_PATH - 1] = 0;
                break;

			case L'S':
				if (!argv[i][2])
				{
					if (++i >= argc)
					{
						fprintf(stderr, "ERROR: Invalid argument for %S\n\n", argv[i-1]);
						displaySyntax();
		                exit(1);
					}

					sptr = argv[i];
				}
				else
					sptr = &argv[i][2];

			    wcsncpy(aSpecialBuild, sptr, MAX_PATH);
				aSpecialBuild[MAX_PATH - 1] = 0;
                break;

			case L'T':
				if (!argv[i][2])
				{
					if (++i >= argc)
					{
						fprintf(stderr, "ERROR: Invalid argument for %S\n\n", argv[i-1]);
						displaySyntax();
		                exit(1);
					}

					sptr = argv[i];
				}
				else
					sptr = &argv[i][2];

			    wcsncpy(aTrademark, sptr, MAX_PATH);
				aTrademark[MAX_PATH - 1] = 0;
                break;

			case L'V':
				if (!argv[i][2])
				{
					if (++i >= argc)
					{
						fprintf(stderr, "ERROR: Invalid argument for %S\n\n", argv[i-1]);
						displaySyntax();
		                exit(1);
					}

					sptr = argv[i];
				}
				else
					sptr = &argv[i][2];

			    wcsncpy(aFileVersion, sptr, MAX_PATH);
				aFileVersion[MAX_PATH - 1] = 0;
                break;

			case L'v':
				if (!argv[i][2])
				{
					if (++i >= argc)
					{
						fprintf(stderr, "ERROR: Invalid argument for %S\n\n", argv[i-1]);
						displaySyntax();
		                exit(1);
					}

					sptr = argv[i];
				}
				else
					sptr = &argv[i][2];

			    wcsncpy(aProductVersion, sptr, MAX_PATH);
				aProductVersion[MAX_PATH - 1] = 0;
                break;

			case L'H':
			case L'h':
			case L'?':
				displaySyntax();
		        exit(0);
                break;

			default:
				displaySyntax();
		        exit(1);
				break;
			}
		}
		else
		{
			wcsncpy(TargetName, argv[i], MAX_PATH);
			TargetName[MAX_PATH - 1] = 0;

			break;
		}
	}
  }


void diagArgs(void)
{
	  if (*aCompanyName)
	      fprintf(stderr, "\tCompany Name = %S\n", aCompanyName);
	  if (*aCopyright)
	      fprintf(stderr, "\tCopyright = %S\n", aCopyright);
	  if (*aFileDescription)
	      fprintf(stderr, "\tFile Description = %S\n", aFileDescription);
	  if (*aInternalFileName)
	      fprintf(stderr, "\tInternal File Name = %S\n", aInternalFileName);
	  if (*aOriginalFileName)
	      fprintf(stderr, "\tOriginal File Name = %S\n", aOriginalFileName);
	  if (*aProductName)
	      fprintf(stderr, "\tProduct Name = %S\n", aProductName);
	  if (*aPrivateBuild)
	      fprintf(stderr, "\tPrivate Build = %S\n", aPrivateBuild);
	  if (*aComments)
	      fprintf(stderr, "\tComments = %S\n", aComments);
	  if (*aSpecialBuild)
	      fprintf(stderr, "\tSpecial Build = %S\n", aSpecialBuild);
	  if (*aTrademark)
	      fprintf(stderr, "\tTrademark = %S\n", aTrademark);
	  if (*aFileVersion)
	      fprintf(stderr, "\tFile Version = %S\n", aFileVersion);
	  if (*aProductVersion)
	      fprintf(stderr, "\tProduct Version = %S\n", aProductVersion);
}



int commasToDots(wchar_t *aString)
{
	int count = 0;

	while(*aString)
	{
		if (*aString == L',')
		{
			count++;
			*aString = L'.';
		}
		else
			if (*aString == L'.')
				count++;

		aString++;
	}

	return(count);
}



/* Modify copies of the FileVersion and ProductVersion strings and set  */
/* the binary version fields appropriately.                             */
int formatVersions(void)
{
	wchar_t tempVerString[MAX_PATH];
	union    v
	{
		unsigned long  longs[2];
		unsigned short shorts[4];
	} v;


    if (*aFileVersion)
	{
		v.longs[0] = 0;
		v.longs[1] = 0;
		wcscpy(tempVerString, aFileVersion);
		commasToDots(tempVerString);

		if (swscanf(tempVerString, L"%hd.%hd.%hd.%hd", &v.shorts[1], &v.shorts[0],
			                                      &v.shorts[3], &v.shorts[2]) < 2)
		{
			fprintf(stderr, "ERROR: Invalid FileVersion format\n");
			displaySyntax();
			exit(1);
		}

		/* Assign the binary version values */
		if (VersionBuffer.ValueLength)
		{
		    pFixedInfo->FileVersionHi = v.longs[0];
		    pFixedInfo->FileVersionLow = v.longs[1];
		}
	}

	if (*aProductVersion)
	{
		v.longs[0] = 0;
		v.longs[1] = 0;
		wcscpy(tempVerString, aProductVersion);
		commasToDots(tempVerString);

		if (swscanf(tempVerString, TEXT("%hd.%hd.%hd.%hd"), &v.shorts[1], &v.shorts[0],
			                                              &v.shorts[3], &v.shorts[2]) < 2)
		{
			fprintf(stderr, "ERROR: Invalid ProductVersion format\n");
			displaySyntax();
			exit(1);
		}

		if (VersionBuffer.ValueLength)
		{
		    pFixedInfo->ProductVersionHi = v.longs[0];
		    pFixedInfo->ProductVersionLow = v.longs[1];
		}
	}

	return(0);
}



int wmain(int argc, wchar_t **argv)
  {
  int   Size;
  DWORD myError;
  WORD  myLanguage = 0;
  HANDLE hUpdate;

  getArgs(argc, argv);

  if (gDiag)
  {
	  fprintf(stderr, "The following variables were set.\n");
	  diagArgs();
  }

  if (!*TargetName)
  {
	  fprintf(stderr, "ERROR: TargetFile filename not set\n");
	  displaySyntax();
	  exit(1);
  }

  if (gDiag)
	  fprintf(stderr, "Target filename = %S.\n", TargetName);

  if ((TargetFile = LoadLibraryEx(TargetName, NULL, LOAD_LIBRARY_AS_DATAFILE)) == NULL)
  {
	  fprintf(stderr, "ERROR: Cannot LoadLibrary for %S\n", TargetName);
	  exit(1);
  }

  FreeLibrary(TargetFile);

  if (gDiag)
	  fprintf(stderr, "\t%S exists and is a loadable file type.\n", TargetName);

  if ((Size = checkVersionInfo(TargetName)) > 0)
  {
    gVerExists = 1;

    if (gDiag)
	  fprintf(stderr, "\t%S already has a VERSIONINFO resource.\n", TargetName);

    if (Size > sizeof(VersionBuffer))
    {
       fprintf(stderr, "VERSIONINFO is larger than VersionBuffer (%d)\n", Size);
       exit(1);
    }

    if (gDiag)
	  fprintf(stderr, "\t\tBuffer size is large enough.\n");

    if ((Size = versionInfoCopy(TargetName, &VersionBuffer, Size)) == 0)
    {
       fprintf(stderr, "ERROR: Couldn't get version info.\n");
       exit(1);
    }

    if (gDiag)
	  fprintf(stderr, "\t\tVERSIONINFO from %S copied successfully to buffer.\n", TargetName);
  }
  else
  {
    if (gDiag)
	  fprintf(stderr, "\t%S currently has no VERSIONINFO resource.\n", TargetName);

    if (!resourceSectionExists(TargetName))
	{
		if (gDiag)
			fprintf(stderr, "\t%S currently has no resource section.\n", TargetName);

		if (!addResourceSection(TargetName))
		{
			fprintf(stderr, "Cannot modify %S. No .rsrc section exists.\n", TargetName);
			exit(50);
		}
	}

	if ((Size = checkVersionInfo(MyName)) == 0)
    {
       fprintf(stderr, "%S\n", MyName);
       fprintf(stderr, "ERROR: Couldn't find my own version info.\n");
       exit(1);
    }

    if (gDiag)
	  fprintf(stderr, "\tEmbedded VERSIONINFO resource is available.\n");

    if (Size > sizeof(VersionBuffer))
    {
       fprintf(stderr, "VERSIONINFO is larger than VersionBuffer %d\n", Size);
       exit(1);
    }

    if (gDiag)
	  fprintf(stderr, "\t\tBuffer size is OK.\n");

    if ((Size = versionInfoCopy(MyName, &VersionBuffer, Size)) == 0)
    {
       fprintf(stderr, "ERROR: Couldn't get version info.\n");
       exit(1);
    }

    if (gDiag)
	  fprintf(stderr, "\t\tEmbedded VERSIONINFO copied successfully to buffer.\n");
  }

  if (VersionBuffer.ValueLength)
	  pFixedInfo = (FixedInfo *)&VersionBuffer.Buffer[0];

  formatVersions();

  if (!gVerExists && !*aOriginalFileName)
	  GetFullPathName(TargetName, MAX_PATH, fullTargetPath, &targetFileName);

  if (setVersionStringInfoBuff() != 0)
  {
	  fprintf(stderr, "\t\tERROR: No StringFileInfo block in VERSIONINFO. Can't modify.\n");
	  exit(1);
  }

  if (swscanf(stringInfo->StringTableKey, L"%4hx", &myLanguage) != 1)
	  myLanguage = MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US);

  versionInfoUpdate();

  if ((hUpdate = BeginUpdateResource(TargetName, FALSE)) == NULL)
  {
	  fprintf(stderr, "ERROR: Cannot get resource update handle to %S\n", TargetName);
	  fprintf(stderr, "GetLastError() returned: %d\n", GetLastError());
	  exit(1);
  }

  if (UpdateResource(hUpdate, RT_VERSION, MAKEINTRESOURCE(1), myLanguage,
					&VersionBuffer, VersionBuffer.ResourceLength) == FALSE)
  {
	  myError = GetLastError();
	  fprintf(stderr, "ERROR: UpdateResource(), error code = %u\n", myError);
	  exit(1);
  }

  if (EndUpdateResource(hUpdate, FALSE) == FALSE)
  {
	  myError = GetLastError();
	  fprintf(stderr, "ERROR: EndUpdateResource(), error code = %u\n", myError);
	  exit(1);
  }

  return(0);
  }

