/* verinfo.c - VERSIONINFO checking routines. No files are harmed in this code! */
#define UNICODE

#include <windows.h>
#include <winver.h>
#include <stdio.h>
#include <string.h>
#include "verinfo.h"

extern wchar_t aCompanyName[MAX_PATH];
extern wchar_t aCopyright[MAX_PATH];
extern wchar_t aCompatibilityLevel[MAX_PATH];
extern wchar_t aFileDescription[MAX_PATH];
extern wchar_t aInternalFileName[MAX_PATH];
extern wchar_t aOriginalFileName[MAX_PATH];
extern wchar_t aProductName[MAX_PATH];
extern wchar_t aPrivateBuild[MAX_PATH];
extern wchar_t aComments[MAX_PATH];
extern wchar_t aSpecialBuild[MAX_PATH];
extern wchar_t aTrademark[MAX_PATH];
extern wchar_t aFileVersion[MAX_PATH];
extern wchar_t aProductVersion[MAX_PATH];

extern wchar_t *targetFileName;

extern VerBuff VersionBuffer;
extern StringInfoBuff *stringInfo;
extern int gVerExists;
extern int gDiag;

typedef struct stringKey
{
WORD len;
WORD valueLen;
WORD type;
wchar_t key[1];
} stringKey;


int checkVersionInfo(wchar_t *fileName)
{
	DWORD junkHandle;

	return(GetFileVersionInfoSize(fileName, &junkHandle));
}

int versionInfoCopy(wchar_t *fileName, void * versionInfoBuffer, int bufferSize)
{
	int test;

	if ((test = checkVersionInfo(fileName)) == 0 || test > bufferSize)
		return(0);

	if (!GetFileVersionInfo(fileName, 0, bufferSize, versionInfoBuffer))
		return(0);

	return((int) *(unsigned short *)versionInfoBuffer);
}

static int versionInfoStringUpdate(wchar_t *key, wchar_t *newValue)
{
	int i;
	int align;
	int found = 0;
	int keyBytes;
	int oldValBytes;
	int newValBytes;
	int difference;
	unsigned short j = 0;
	char *Buffer;
	char *buffLimit;
	char *endTable;
	char *endResource;
	wchar_t *value;
	stringKey *psKey;

	Buffer = (char *)stringInfo->Buffer;
	endTable = (char *)&stringInfo->StringTableLen + stringInfo->StringTableLen - 1;
	endResource = (char *)&VersionBuffer + VersionBuffer.ResourceLength - 1;
	buffLimit = (char *)&VersionBuffer + sizeof(VersionBuffer) - 1;
	psKey = (stringKey *)&Buffer[0];

	for (i = 0; &Buffer[i] < endTable; i += j)
	{
		psKey = (stringKey *)&Buffer[i];
        j = *(unsigned short *)&Buffer[i];

		if (align = j % 4)
			j += 4 - align;

		if (!wcscmp(psKey->key, key))
		{
			found = 1;

			if (gDiag && gVerExists)
				fprintf(stderr, "\tFound key %S\n", psKey->key);

			break;
		}
	}

	if (!found)
	{
		if (gDiag)
			fprintf(stderr, "\tInserting an empty %S string\n", key);

		psKey = (stringKey *)&Buffer[i];

		difference = sizeof(stringKey) + (wcslen(key) * 2);
		
		if (align = difference % 4)
			difference += 4 - align;

		if (endResource + difference > buffLimit)
		{
		    fprintf(stderr, "\n\nERROR: Buffer not large enough for new string\n\n");
		    exit(99);
		}

		memmove((char *)psKey + difference, psKey, ((char *)endResource - (char *)psKey) + 1);
		
		wcscpy(psKey->key, key);
		psKey->type = 1;
		psKey->valueLen = 0;
		psKey->len = difference;

		endResource += difference;
		endTable += difference;
		stringInfo->StringTableLen += difference;
		stringInfo->StringFileInfoLen += difference;
		VersionBuffer.ResourceLength += difference;
	}


	keyBytes = wcslen(psKey->key) * 2 + 2;

	if ((keyBytes + 2) % 4)
		keyBytes += 2;

	value = (wchar_t *)((char *)psKey->key + keyBytes);

	if (psKey->valueLen)
	    oldValBytes = wcslen(value) * 2 + 2;
	else
		oldValBytes = 0;

	if (oldValBytes % 4)
		oldValBytes += 2;

	newValBytes = wcslen(newValue) * 2 + 2;

	if (newValBytes % 4)
		newValBytes += 2;

	difference = newValBytes - oldValBytes;

	if (VersionBuffer.ResourceLength + difference > sizeof(VersionBuffer))
	{
		fprintf(stderr, "\n\nERROR: Buffer not large enough\n\n");
		exit(99);
	}

	if (gDiag)
	{
        fprintf(stderr, "Setting %S\n", psKey->key);

		if (psKey->valueLen)
			fprintf(stderr, "\tOld value = \"%S\"\n", value);
		else
			fprintf(stderr, "\tOld value = \"\"\n");
	}

	if (difference)
	{
		memmove((char *)value + newValBytes, (char *)value + oldValBytes,
		         endResource - (char *)value + oldValBytes + 1);
		endResource += difference;
		psKey->valueLen = wcslen(newValue) + 1;
		psKey->len += difference;
		stringInfo->StringTableLen += difference;
		stringInfo->StringFileInfoLen += difference;
		VersionBuffer.ResourceLength += difference;
	}


    wcsncpy(value, newValue, newValBytes / 2);

	if (gDiag)
		fprintf(stderr, "\tNew value = \"%S\"\n", value);

	return(difference);
}


int versionInfoUpdate(void)
{
	  if (*aCompanyName)
          versionInfoStringUpdate(TEXT("CompanyName"), aCompanyName);

	  if (*aCopyright)
          versionInfoStringUpdate(TEXT("LegalCopyright"), aCopyright);

	  if (*aCompatibilityLevel)
		  versionInfoStringUpdate(TEXT("CompatibilityLevel"), aCompatibilityLevel);

	  if (*aFileDescription)
          versionInfoStringUpdate(TEXT("FileDescription"), aFileDescription);

	  if (*aInternalFileName)
          versionInfoStringUpdate(TEXT("InternalName"), aInternalFileName);

	  if (*aOriginalFileName)
          versionInfoStringUpdate(TEXT("OriginalFilename"), aOriginalFileName);
	  else
		  if (targetFileName && *targetFileName)
              versionInfoStringUpdate(TEXT("OriginalFilename"), targetFileName);

	  if (*aProductName)
          versionInfoStringUpdate(TEXT("ProductName"), aProductName);

	  if (*aPrivateBuild)
          versionInfoStringUpdate(TEXT("PrivateBuild"), aPrivateBuild);

	  if (*aComments)
          versionInfoStringUpdate(TEXT("Comments"), aComments);

	  if (*aSpecialBuild)
          versionInfoStringUpdate(TEXT("SpecialBuild"), aSpecialBuild);

	  if (*aTrademark)
          versionInfoStringUpdate(TEXT("LegalTrademarks"), aTrademark);

	  if (*aFileVersion)
          versionInfoStringUpdate(TEXT("FileVersion"), aFileVersion);

	  if (*aProductVersion)
          versionInfoStringUpdate(TEXT("ProductVersion"), aProductVersion);

	  return(0);
}


int setVersionStringInfoBuff(void)
{
	int   align;
	char *lastByte;

	if (VersionBuffer.ValueLength)
	{
	    stringInfo = (StringInfoBuff *)(&VersionBuffer.Buffer[0] + sizeof(FixedInfo));
	}
	else
	{
		if (gDiag)
			fprintf(stderr, "Note: No FIXEDFILEINFO available.\n");

		stringInfo = (StringInfoBuff *)&VersionBuffer.Buffer[0];
	}

	lastByte = (char *)&VersionBuffer + VersionBuffer.ResourceLength - 1;

	if (!wcscmp(L"StringFileInfo", stringInfo->StringFileInfoKey))
	{
		if ((char *)stringInfo + stringInfo->StringFileInfoLen <= lastByte)
		    return(0);
	}

	if (!wcscmp(L"VarFileInfo", stringInfo->StringFileInfoKey))
	{
		stringInfo = (StringInfoBuff *) ((char *)stringInfo + stringInfo->StringFileInfoLen);

		if (align = (long)stringInfo % 4)
			(char *)stringInfo += align % 4;

	    if (!wcscmp(L"StringFileInfo", stringInfo->StringFileInfoKey))
		{
		    if ((char *)stringInfo + stringInfo->StringFileInfoLen <= lastByte + 1)
		        return(0);
		}
	}

	return(1);
}


